//////////////////////////////////////////////////////////////////////////////////////////////
// Agent.cpp
// The file defines the function GetAgentSet.
// The function GetAgentSet creates connectivity among the cells (precursors and osteocytes).
// This function is called by main.cpp. 
// Created by Jitendra Prasad, Ph.D. / Brandon J. Ausk / Sundar Srinivasan, Ph.D.
// University of Washington at Seattle. 
//////////////////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <cmath>
using namespace std;

vector<vector<short> > GetAgentSet(int ncell, int nob, double rad, vector<vector<double> > cellXY)
{
	vector<short> VecShort(ncell,0);
	vector<vector<short> > agentset(ncell,VecShort);
	for (int i=0; i<ncell; i++)
		for (int j=0; j<ncell; j++)
		{
			double dist;
			if (i<nob && j<nob)
				dist = 0;
			else
				dist = sqrt( pow(cellXY[i][0]-cellXY[j][0],2) + pow(cellXY[i][1]-cellXY[j][1],2) );
			if (dist > 0 && dist < rad)
				agentset[i][j] = 1;
		}

	for (int i=1; i<nob/2-1; i++)
	{
		agentset[i][i-1] = 1;
		agentset[i][i+1] = 1;
		agentset[i+nob/2][i+nob/2-1] = 1;
		agentset[i+nob/2][i+nob/2+1] = 1;
	}
	// osteoblast connectivities for ob number 1 and nob/2 on endocortical surface
	agentset[0][1] = 1;
	agentset[0][nob/2-1] = 1;
	agentset[nob/2-1][0] = 1;
	agentset[nob/2-1][nob/2-2] = 1;
	// osteoblast connectivities for ob number nob/2+1 and nob on periosteal surface
	agentset[nob/2][nob/2+1]=1;
	agentset[nob/2][nob-1]=1;
	agentset[nob-1][nob/2]=1;
	agentset[nob-1][nob-2]=1;
	return agentset;
}
