//////////////////////////////////////////////////////////////////////////////////////////////
// Iterate.cpp
// The file defines the function DoIterations.
// The function DoIterations computes Relative Periosteal Bone Formation Rate (rp.BFR).
// This function is called by EvalObjFun.cpp. 
// Created by Jitendra Prasad, Ph.D. / Brandon J. Ausk / Sundar Srinivasan, Ph.D.
// University of Washington at Seattle. 
//////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <vector>		
#include <algorithm>
#include <cmath>

//#include <fstream>
//#include <sstream>
//#include <string>
//#include <cstring>

using namespace std; 
#include "Output.h"

vector<vector<double> > DoIterations(float Tcmin, vector<vector<double> > ParaList, int nPara, int nABM, 
									 vector<vector<double> > Prot, int nProt, double extratime, 
									 vector<double> prottime, vector<double> maxtime,
									 int nob, int ncell, 
									 vector<vector<int> > agentConnection, vector<int> agentcount, 
									 vector<int> dummyagentcount, vector<vector<double> > strain, bool IsYoung)
{
	vector<double> Vec1(nProt,0);
	vector<vector<double> > rpBFR_Para(nPara,Vec1);
	Vec1.clear();
	
	double temax, temin, alphae, CaStore0, Rt, alphaStore, Decay, deNFATThreshold, nucNFATMax, MARmax;
	for (int ipara = 0; ipara < nPara; ipara++) //usually there is only one Para while evaluating Optimization Objective Function
	{
		
		temax = ParaList[ipara][0];
		temin = 0;
		if (temin >= temax) temin = 0.9999*temax;
		
		Rt = ParaList[ipara][1];
		CaStore0 = ParaList[ipara][2];
		alphae = (temax-temin)/10.0;
		alphaStore = CaStore0/10.0;
		
		Decay = ParaList[ipara][3];
		
		deNFATThreshold = 0; //for simulated annealing
		
		nucNFATMax = ParaList[ipara][4];
		MARmax = ParaList[ipara][5];
		
		vector<double> Vec2(nob,0), CaStrain(ncell,0);
		vector<vector<double> > MARCell(nProt, Vec2);
		
		int nMaxMaxtime = int (*(max_element(maxtime.begin(),maxtime.end()))); //Finding the maximum value
		
		vector<vector<double> > Mat1(nMaxMaxtime,Vec2);
		vector<vector<vector<double> > > Realtime(nProt,Mat1); 
		Mat1.clear();
		
		//  Protocal Loop
		
		for (int iprot=0; iprot<nProt; iprot++)
		{
			
			for(int j=nob; j<ncell; j++)
				CaStrain[j] = 100.0/(1.+exp(-(strain[iprot][j]-
											  (temax + temin)/2.0)/alphae)); 
			//Ocyte is the mechanosensor
			
			//  Initiating for storage purposes
			vector<vector<double> > CaCell(int(maxtime[iprot]),nob);
			vector<vector<double> > deNFAT(int(maxtime[iprot]+extratime/2),Vec2); //zero initialized
			vector<double> Energy(ncell,CaStore0);
			vector<double> CaComm(ncell,0);  //  Initialize communication equal to zero for t=1
			vector<double> CaActual(ncell,0); //  This will be fed into CaCell and used for communication
			
			// NOTE : The first index it CaCell is t=1; That means there is no t=0
			int keystim = 0;     //   If Keystim equals 1 stimulus is OFF 
			//If Keystim equals 0 stimulus is ON
			int keycycle = 0; //check data type    //   Keycycle used to track cycle groupin
			
			// First Time Loop is for Strain portion
			// Second Time Loop is for Communication during 'Extra Time'
			
			for(int t=0; t<prottime[iprot]; t++)
			{
				//replenishing the stores if the cell is quiet
				for (int i=0; i<ncell; i++) 
				{
					if (CaActual[i]<=1.0e-06)
						Energy[i] = Energy[i]+Rt*1.0/(1.0+exp((Energy[i]-CaStore0/2.0)/alphaStore)); //energy store at i = 0 corresponds to t = 1
					if (Energy[i] > CaStore0)
						Energy[i] = CaStore0;
				}
				//keystim equal to 0 means stimulus on
				if (keystim == 0)
				{                
					//new communication try
					for (int i=0; i<ncell; i++)
					{
						double product = 0;
						for (int k=0; k < agentcount[i]; k++)
							product = product + CaActual[agentConnection[i][k]]; 
						CaComm[i]=product/double(dummyagentcount[i]); 
						if (CaComm[i]<= Tcmin)
							CaComm[i] = 0;
					}
					for (int i=0; i<ncell; i++)
					{
						// This sets CaActual to the higher value between strain and
						// communication induced calcium
						CaActual[i]=max(CaStrain[i],CaComm[i]);	
						
						// This sets CaActual to the minimum between the desired
						// response and what is allowable due to energy levels
						CaActual[i]=min(CaActual[i],Energy[i]);
					}
				}
				else
				{
					for (int i=0; i<ncell; i++)
					{
						double product = 0;
						for (int k=0; k < agentcount[i]; k++)
							product = product + CaActual[agentConnection[i][k]]; 
						CaComm[i]=product/double(dummyagentcount[i]); 
						if (CaComm[i]<= Tcmin)
							CaComm[i] = 0;
					}
					// This sets CaActual to the minimum between the desired
					// response and what is allowable due to energy levels
					for (int i=0; i<ncell; i++)
						CaActual[i]=min(CaComm[i],Energy[i]);
				}
				
				// Setting small signals equal to zero%%%%%
				for (int i=0; i<ncell; i++)
				{
					if (CaActual[i] < 0.67)
						CaActual[i] = 0.0; // correction for the sigmoid value when under min threshold
					// value of 100*1/(1+exp(5))
					// Depleting energy due to response
					Energy[i] = Energy[i] - CaActual[i];
					if ( Energy[i] < 0 )
						Energy[i] = 0;
				}
				
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				// Used to track when stimulus is applied %%%%
				keycycle = keycycle + 1;
				if (keycycle == int(Prot[iprot][3]))
					keystim = 1;
				if (keycycle == int(Prot[iprot][2] + Prot[iprot][3]))
				{
					keystim = 0;
					keycycle = 0;
				}
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
				
				for(int i=0; i<nob; i++)
				{
					CaCell[t][i]=CaActual[i]; //OPT: Eliminate CaCell; use CaActual instead
				}
				if (t>0)
					for(int i=0; i<nob; i++)
						deNFAT[t][i]=Decay*deNFAT[t-1][i]+CaCell[t][i]; //Computing the gene transcription
				//occuring in the real-time based upon geometrically weighted moving
				//averages with an infinite backward window
				
				for(int i=0; i<nob; i++)
					Realtime[iprot][t][i]=CaActual[i];
			}
			
			//%%%%% Extra Time %%%%%%
			for(int t = int(prottime[iprot]); t < int(maxtime[iprot]); t++)
			{
				// replenishing the stores if the cell is quiet
				for (int i=0; i<ncell; i++)
				{
					if (CaActual[i]<=1.0e-06)
						Energy[i] = Energy[i]+Rt*1.0/(1+exp((Energy[i]-CaStore0/2.0)/alphaStore));
					
					if (Energy[i] > CaStore0)
						Energy[i] = CaStore0;
				}
				
				for (int i=0; i<ncell; i++)
				{
					double product = 0;
					for (int k=0; k < agentcount[i]; k++)
						product = product + CaActual[agentConnection[i][k]]; 
					CaComm[i]=product/double(dummyagentcount[i]); 
					if (CaComm[i]<= Tcmin)
						CaComm[i] = 0;
				}
				// This sets CaActual to the minimum between the desired
				// response and what is allowable due to energy levels
				for (int i=0; i<ncell; i++)
					CaActual[i]=min(CaComm[i],Energy[i]);
				
				//%%%%%% Setting small signals equal to zero%%%%%
				for (int i=0; i<ncell; i++)
				{
					if (CaActual[i] < 0.67)
						CaActual[i] = 0.0; // correction for the sigmoid value when under min threshold
					
					// Depleting energy due to response
					Energy[i] = Energy[i] - CaActual[i];
					if ( Energy[i] < 0 )
						Energy[i] = 0;
				}
				
				for(int i=0; i<nob; i++)
					CaCell[t][i]=CaActual[i]; 
				
				for(int i=0; i<nob; i++)
					deNFAT[t][i]=Decay*deNFAT[t-1][i]+CaCell[t][i]; //Computing the gene transcription
				//occuring in the real-time based upon geometrically weighted moving
				//averages with an infinite backward window
				
				for(int i=0; i<nob; i++)
					Realtime[iprot][t][i]=CaActual[i];
			}
			
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//%Computing Gene transcription for a bit after the protocol ends
			//%for all but the largest protocol (typically rest-inserted,
			//%which really does not require any more computation than that
			//%allowed by the 'extratime' number...
			
			int endtime = nMaxMaxtime;
			if (int(maxtime[iprot]) < nMaxMaxtime)
			{		
				endtime = int(maxtime[iprot] + extratime/2);
			}
			
			for (int t = int(maxtime[iprot]); t < endtime; t++) // CaCell for t > maxtime (i.e. prottime+extratime) is forced to be zero
			{
				for(int i=0; i<nob; i++)
					deNFAT[t][i]=Decay*deNFAT[t-1][i]; //since CaCell[t][i] = 0 here
				//%Computing the gene transcription
				//%occuring in the real-time based upon geometrically weighted moving averages with an infinite backward window
			}
			
			//Printing Ca2+ in Periosteal Osteoblasts For Each Protocol
			OutputToFile("Ca", IsYoung, iprot, Realtime[iprot], "Time(s)", "Osteoblast_No.");
			
			//Printing NFAT in Periosteal Osteoblasts For Each Protocol
			OutputToFile("NFAT", IsYoung, iprot, deNFAT, "Time(s)", "Osteoblast_No.");		
			
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			
			vector<double> nucNFAT(nob,0);
			for(int i=0; i<nob; i++)
			{
				double deNFATSum = 0;
				for(int t=0; t < endtime; t++)
				{
					deNFAT[t][i] = deNFAT[t][i] - deNFATThreshold;
					if(deNFAT[t][i] < 0 )
						deNFAT[t][i] = 0;
					deNFATSum = deNFATSum + deNFAT[t][i];
				}
				nucNFAT[i] = deNFATSum;
				if (nucNFAT[i] > 0)
					MARCell[iprot][i] = MARmax/(1.0+exp(-(nucNFAT[i]-nucNFATMax/2.0)/0.1/nucNFATMax));
			}
		}
		
		//Printing MAR in Periosteal Osteoblasts For Each Protocol
		OutputToFile("rp.MAR", IsYoung, -1, MARCell, "Prot._No.", "Osteoblast_No.");		
		
		vector<double> BFRPeri(nProt,0); // periosteal tissue level BFR
		vector<double> BFREndo(nProt,0); // endocortical tissue level BFR
		vector<double> MARPeri(nProt,0); // periosteal tissue level MAR
		vector<double> MAREndo(nProt,0); // endocortical tissue level MAR
		vector<double> MSPeri(nProt,0); // periosteal tissue level MS
		vector<double> MSEndo(nProt,0); // endocortical tissue level MS
		
		for (int iprot=0; iprot < nProt; iprot++) 		
		{
			int nNonZeroMARCellPeri = 0;
			int nNonZeroMARCellEndo = 0;
			double sumMARCellPeri = 0;
			double sumMARCellEndo = 0;
			
			for(int i=0; i < nob/2; i++)
			{
				if(MARCell[iprot][i] != 0)
					nNonZeroMARCellEndo = nNonZeroMARCellEndo + 1;
				sumMARCellEndo = sumMARCellEndo + MARCell[iprot][i];
			}
			for(int i=nob/2; i < nob; i++)
			{
				if(MARCell[iprot][i] != 0)
					nNonZeroMARCellPeri = nNonZeroMARCellPeri + 1;
				sumMARCellPeri = sumMARCellPeri + MARCell[iprot][i];
			}
			if(nNonZeroMARCellPeri != 0)
			{
				MSPeri[iprot] = double(nNonZeroMARCellPeri)/double(nob)*2.0*100.0; // MS is simply the no of cells that have nonzero MAR
				MARPeri[iprot] = sumMARCellPeri/double(nNonZeroMARCellPeri); // MAR is the non-zero ave of MAR in each cell
				BFRPeri[iprot] = MSPeri[iprot]*MARPeri[iprot]/100.0;
			}
			if(nNonZeroMARCellEndo != 0)
			{
				MSEndo[iprot] = double(nNonZeroMARCellEndo)/double(nob)*2.0*100.0; // MS is simply the no of cells that have nonzero MAR
				MAREndo[iprot] = sumMARCellEndo/double(nNonZeroMARCellEndo); // MAR is the non-zero ave of MAR in each cell
				BFREndo[iprot] = MSEndo[iprot]*MAREndo[iprot]/100.0;
			}
			rpBFR_Para[ipara][iprot] = BFRPeri[iprot]; 
		}
		//Printing BFR in Periosteal Osteoblasts For Each Protocol
		OutputToFile("rp.BFR", IsYoung, -2, rpBFR_Para, "Param._No.", "Prot._No.");	
	}
	return rpBFR_Para;
}
