/////////////////////////////////////////////////////////////////////////////////////////////
// main.cpp
// This is the main file of the program. 
// Other subroutines are called from here.
// Created by Jitendra Prasad, Ph.D. / Brandon J. Ausk / Sundar Srinivasan, Ph.D.
// University of Washington at Seattle.
//////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream> //main c++ header
#include <fstream> //header for file input/output
#include <vector> //vector data structure
#include <cmath> //Math routines
#include <ctime> //Time routines
#include <algorithm> //algorithms for vector manipulation
#include <sstream> //string input/output manipulation
#include <string> //string manipulation
using namespace std;

#include "Input.h" //header to read input file
#include "Agent.h" //To creat agent connectivity
#include "EvalObjFun.h" //To computate rp.BFR

int main()
{
	cout << "Wait ..." << endl;
	double extratime = 200; 
	//cout << "how many extra seconds should real-time signaling be computed for? :"; 
	//cin >> extratime; //read extra time (e.g. 200)
	
	double rad = 5; 
	//cout << "what is the canalicular length? :";
	//cin >> rad; //read canalicular length (e.g. 5)
	
	float Tcmin = 2.5;
	//cout << "what is Tcmin (intercellular Ca++ communication threshold)? :";
	//cin >> Tcmin; //read (e.g. 2.5)
	
	// File Input and Calculations of Intermediate Variables
	
	// ***************************************************
	// ************** YOUNG MOUSE INPUT *****************
	// ***************************************************
	
	//Reading the protocols
	vector<vector<double> > Prot_young = ReadDoubleMatrix("prot_young.txt");
	int nProt_young = int(Prot_young.size());
	
	vector<double> prottime_young(nProt_young), maxtime_young(nProt_young);
	for (int i = 0; i < nProt_young; i++)
	{
		prottime_young[i] = ( Prot_young[i][3] + Prot_young[i][2] ) * Prot_young[i][1] / Prot_young[i][3];
		maxtime_young[i] = prottime_young[i] + extratime;
	}
	
	// Setting up section properties YOUNG
	double Ix_young = 4024620.0;
	double Iy_young = 3278040.0;
	double Ixy_young = -202161.0;
	double Ixyx_young = Ix_young*Iy_young-pow(Ixy_young,2);
	double Area_young = 4239.0;
	
    //Calculating normal force and bending moment YOUNG
	vector<double> Fz_young(nProt_young), Mx_young(nProt_young), My_young(nProt_young);
	for (int i = 0; i < nProt_young; i++)
	{
		Fz_young[i] = 0.736251/1600.0*Prot_young[i][0];
		Mx_young[i] = 23.765/1600.0*Prot_young[i][0];
		My_young[i] = -112.079/1600.0*Prot_young[i][0];
	}
	
	// Reading Osteoblast Coordinates
	vector<vector<double> > obXY_young = ReadDoubleMatrix("precursor_young.txt");
	int nob_young = int(obXY_young.size());
	
	// Reading Osteocytes Coordinates
	vector<vector<double> > ocXY_young = ReadDoubleMatrix("ocytes_young.txt");
	int noc_young = int(ocXY_young.size());
	
	// combined cell coordinates
	int ncell_young = nob_young+noc_young;
	vector<vector<double> > cellXY_young = obXY_young;
	cellXY_young.insert(cellXY_young.end(),ocXY_young.begin(),ocXY_young.end());
	
	vector<vector<short> > agentset_young = GetAgentSet(ncell_young, nob_young, rad, cellXY_young);
	vector<vector<int> > agentConnection_young(ncell_young);
	
	vector<int> agentcount_young(ncell_young), dummyagentcount_young(ncell_young);
	for(int i=0; i<ncell_young; i++)
	{
		int sum = 0;
		for(int j=0; j<ncell_young; j++)
			if(agentset_young[i][j]==1)
			{
				agentConnection_young[i].push_back(j);
				sum = sum + 1; 
			}
		agentcount_young[i] = sum; // Number of connections for each cell
		if ( sum > 0 )
			dummyagentcount_young[i] = sum;
		else
			dummyagentcount_young[i] = 1; // This steps allows for calculation of cells w/out neighbors
	}
	agentset_young.clear();
	
	//determining strains and strain induced Ca responses in OSTEOCYTIC
	//cells.  Assumes that osteocytes are mechanosensors and osteoblasts
	//cannot sense mechanical strain.
	
	vector<double> Vec_young(ncell_young,0);
	vector<vector<double> > strain_young(nProt_young,Vec_young);
	for (int iprot=0; iprot<nProt_young; iprot++)
		for(int j=0; j<ncell_young; j++)
			strain_young[iprot][j] = abs( Fz_young[iprot]/Area_young + cellXY_young[j][1]*(Mx_young[iprot]*Iy_young-My_young[iprot]*Ixy_young)/Ixyx_young -
										 cellXY_young[j][0]*(My_young[iprot]*Ix_young-Mx_young[iprot]*Ixy_young)/Ixyx_young)*1.0e+06;
	
	// End of Input Variables and Calculations of Intermediate Variables
	
	// **************************************************
	// ************** AGED MOUSE INPUT *****************
	// **************************************************
	
	// Reading the protocols
	vector<vector<double> > Prot_aged = ReadDoubleMatrix("prot_aged.txt");
	int nProt_aged = int(Prot_aged.size());
	
	vector<double> prottime_aged(nProt_aged), maxtime_aged(nProt_aged);
	for (int i = 0; i < nProt_aged; i++)
	{
		prottime_aged[i] = ( Prot_aged[i][3] + Prot_aged[i][2] ) * Prot_aged[i][1] / Prot_aged[i][3];
		maxtime_aged[i] = prottime_aged[i] + extratime;
	}
	
	// Setting up section properties OLD
	double Ix_aged = 4.02047e+006;
	double Iy_aged = 3.53954e+006;
	double Ixy_aged = 97445.7;
	double Ixyx_aged = Ix_aged*Iy_aged-pow(Ixy_aged,2);
	double Area_aged = 3733.0;
	
	//Calculating normal force and bending moment OLD
	vector<double> Fz_aged(nProt_aged), Mx_aged(nProt_aged), My_aged(nProt_aged);
	for (int i = 0; i < nProt_aged; i++)
	{
		Fz_aged[i] = 0.736251/1620.7522*Prot_aged[i][0];
		Mx_aged[i] = 23.765/1620.7522*Prot_aged[i][0];
		My_aged[i] = -112.079/1620.7522*Prot_aged[i][0];
	}
	
	// Reading Osteoblast Coordinates
	vector<vector<double> > obXY_aged = ReadDoubleMatrix("precursor_aged.txt");
	int nob_aged = int(obXY_aged.size());
	
	// Reading Osteocytes Coordinates
	vector<vector<double> > ocXY_aged = ReadDoubleMatrix("ocytes_aged.txt");
	int noc_aged = int(ocXY_aged.size());
	
	// combined cell coordinates
	int ncell_aged = nob_aged+noc_aged;
	vector<vector<double> > cellXY_aged = obXY_aged;
	cellXY_aged.insert(cellXY_aged.end(),ocXY_aged.begin(),ocXY_aged.end());
	
	vector<vector<short> > agentset_aged = GetAgentSet(ncell_aged, nob_aged, rad, cellXY_aged);
	vector<vector<int> > agentConnection_aged(ncell_aged);
	
	vector<int> agentcount_aged(ncell_aged), dummyagentcount_aged(ncell_aged);
	for(int i=0; i<ncell_aged; i++)
	{
		int sum = 0;
		for(int j=0; j<ncell_aged; j++)
			if(agentset_aged[i][j]==1)
			{
				agentConnection_aged[i].push_back(j);
				sum = sum + 1;
			}
		agentcount_aged[i] = sum; // Number of connections for each cell
		if ( sum > 0 )
			dummyagentcount_aged[i] = sum;
		else
			dummyagentcount_aged[i] = 1; // This steps allows for calculation of cells w/out neighbors
	}
	agentset_aged.clear();
	
	//determining strains and strain induced Ca responses in OSTEOCYTIC
	//cells.  Assumes that osteocytes are mechanosensors and osteoblasts
	//cannot sense mechanical strain.
	
	vector<double> Vec_aged(ncell_aged,0);
	vector<vector<double> > strain_aged(nProt_aged,Vec_aged);
	for (int iprot=0; iprot<nProt_aged; iprot++)
		for(int j=0; j<ncell_aged; j++)
			strain_aged[iprot][j] = abs( Fz_aged[iprot]/Area_aged + cellXY_aged[j][1]*(Mx_aged[iprot]*Iy_aged-My_aged[iprot]*Ixy_aged)/Ixyx_aged -
										cellXY_aged[j][0]*(My_aged[iprot]*Ix_aged-Mx_aged[iprot]*Ixy_aged)/Ixyx_aged)*1.0e+06;
	
	// reading parameter ranges, 
	vector<vector<double> > ParaRange = ReadDoubleMatrix("ParaOptima.txt");
	int nABM = int(ParaRange[0].size()); //No. of parameters in every set e.g. 8
	
	vector<double> Para = ParaRange[0];  // Initial Position of the the parameter search
	vector<vector<double> > BFRinVivo_combined;
	vector<double> std_combined;
	
	// End of Input Variables and Calculations of Intermediate Variables
	
	time_t start, end;
	double dif;
	clock_t starttime, stoptime; 
	double timeused;
	time (&start);
	starttime = clock();

	// determining the MLE by calling the following function.
	EvalObjFun(nABM, Para, Tcmin, Prot_young, nProt_young, extratime, 
									prottime_young, maxtime_young,
									nob_young, ncell_young, 
									agentConnection_young, agentcount_young, 
									dummyagentcount_young, strain_young, 
									Prot_aged, nProt_aged,  
									prottime_aged, maxtime_aged,
									nob_aged, ncell_aged, 
									agentConnection_aged, agentcount_aged, 
									dummyagentcount_aged, strain_aged,
									BFRinVivo_combined, std_combined);
	
	stoptime = clock();
	time (&end);
	timeused = (double(stoptime - starttime))/double(CLOCKS_PER_SEC);
	
	cout << "CPU time used: " << timeused << " seconds" << endl;
	dif = difftime (end,start);
	cout << "Wall-clock time used: "  << dif << " seconds" << endl;
	
}
